/*!
   The Plotters SVG backend.

   The plotters bitmap backend allows you to render images by Plotters into SVG vector graphs.

   See the documentation for [SVGBackend](struct.SVGBackend.html) for more details.
*/
mod svg;

pub use svg::SVGBackend;
