// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

#include "manual.h"
#include <stdio.h>

#define PRINT_CONSTANT(CONSTANT_NAME) \
    printf("%s;", #CONSTANT_NAME); \
    printf(_Generic((CONSTANT_NAME), \
                    char *: "%s", \
                    const char *: "%s", \
                    char: "%c", \
                    signed char: "%hhd", \
                    unsigned char: "%hhu", \
                    short int: "%hd", \
                    unsigned short int: "%hu", \
                    int: "%d", \
                    unsigned int: "%u", \
                    long: "%ld", \
                    unsigned long: "%lu", \
                    long long: "%lld", \
                    unsigned long long: "%llu", \
                    float: "%.6f", \
                    double: "%.6f", \
                    long double: "%ld"), \
           CONSTANT_NAME); \
    printf("\n");

int main() {
    PRINT_CONSTANT((gint) PANGO_ALIGN_CENTER);
    PRINT_CONSTANT((gint) PANGO_ALIGN_LEFT);
    PRINT_CONSTANT((gint) PANGO_ALIGN_RIGHT);
    PRINT_CONSTANT(PANGO_ANALYSIS_FLAG_CENTERED_BASELINE);
    PRINT_CONSTANT(PANGO_ANALYSIS_FLAG_IS_ELLIPSIS);
    PRINT_CONSTANT(PANGO_ANALYSIS_FLAG_NEED_HYPHEN);
    PRINT_CONSTANT((gint) PANGO_ATTR_ABSOLUTE_LINE_HEIGHT);
    PRINT_CONSTANT((gint) PANGO_ATTR_ABSOLUTE_SIZE);
    PRINT_CONSTANT((gint) PANGO_ATTR_ALLOW_BREAKS);
    PRINT_CONSTANT((gint) PANGO_ATTR_BACKGROUND);
    PRINT_CONSTANT((gint) PANGO_ATTR_BACKGROUND_ALPHA);
    PRINT_CONSTANT((gint) PANGO_ATTR_BASELINE_SHIFT);
    PRINT_CONSTANT((gint) PANGO_ATTR_FALLBACK);
    PRINT_CONSTANT((gint) PANGO_ATTR_FAMILY);
    PRINT_CONSTANT((gint) PANGO_ATTR_FONT_DESC);
    PRINT_CONSTANT((gint) PANGO_ATTR_FONT_FEATURES);
    PRINT_CONSTANT((gint) PANGO_ATTR_FONT_SCALE);
    PRINT_CONSTANT((gint) PANGO_ATTR_FOREGROUND);
    PRINT_CONSTANT((gint) PANGO_ATTR_FOREGROUND_ALPHA);
    PRINT_CONSTANT((gint) PANGO_ATTR_GRAVITY);
    PRINT_CONSTANT((gint) PANGO_ATTR_GRAVITY_HINT);
    PRINT_CONSTANT(PANGO_ATTR_INDEX_FROM_TEXT_BEGINNING);
    PRINT_CONSTANT(PANGO_ATTR_INDEX_TO_TEXT_END);
    PRINT_CONSTANT((gint) PANGO_ATTR_INSERT_HYPHENS);
    PRINT_CONSTANT((gint) PANGO_ATTR_INVALID);
    PRINT_CONSTANT((gint) PANGO_ATTR_LANGUAGE);
    PRINT_CONSTANT((gint) PANGO_ATTR_LETTER_SPACING);
    PRINT_CONSTANT((gint) PANGO_ATTR_LINE_HEIGHT);
    PRINT_CONSTANT((gint) PANGO_ATTR_OVERLINE);
    PRINT_CONSTANT((gint) PANGO_ATTR_OVERLINE_COLOR);
    PRINT_CONSTANT((gint) PANGO_ATTR_RISE);
    PRINT_CONSTANT((gint) PANGO_ATTR_SCALE);
    PRINT_CONSTANT((gint) PANGO_ATTR_SENTENCE);
    PRINT_CONSTANT((gint) PANGO_ATTR_SHAPE);
    PRINT_CONSTANT((gint) PANGO_ATTR_SHOW);
    PRINT_CONSTANT((gint) PANGO_ATTR_SIZE);
    PRINT_CONSTANT((gint) PANGO_ATTR_STRETCH);
    PRINT_CONSTANT((gint) PANGO_ATTR_STRIKETHROUGH);
    PRINT_CONSTANT((gint) PANGO_ATTR_STRIKETHROUGH_COLOR);
    PRINT_CONSTANT((gint) PANGO_ATTR_STYLE);
    PRINT_CONSTANT((gint) PANGO_ATTR_TEXT_TRANSFORM);
    PRINT_CONSTANT((gint) PANGO_ATTR_UNDERLINE);
    PRINT_CONSTANT((gint) PANGO_ATTR_UNDERLINE_COLOR);
    PRINT_CONSTANT((gint) PANGO_ATTR_VARIANT);
    PRINT_CONSTANT((gint) PANGO_ATTR_WEIGHT);
    PRINT_CONSTANT((gint) PANGO_ATTR_WORD);
    PRINT_CONSTANT((gint) PANGO_BASELINE_SHIFT_NONE);
    PRINT_CONSTANT((gint) PANGO_BASELINE_SHIFT_SUBSCRIPT);
    PRINT_CONSTANT((gint) PANGO_BASELINE_SHIFT_SUPERSCRIPT);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_AL);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_AN);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_B);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_BN);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_CS);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_EN);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_ES);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_ET);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_FSI);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_L);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_LRE);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_LRI);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_LRO);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_NSM);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_ON);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_PDF);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_PDI);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_R);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_RLE);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_RLI);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_RLO);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_S);
    PRINT_CONSTANT((gint) PANGO_BIDI_TYPE_WS);
    PRINT_CONSTANT((gint) PANGO_COVERAGE_APPROXIMATE);
    PRINT_CONSTANT((gint) PANGO_COVERAGE_EXACT);
    PRINT_CONSTANT((gint) PANGO_COVERAGE_FALLBACK);
    PRINT_CONSTANT((gint) PANGO_COVERAGE_NONE);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_LTR);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_NEUTRAL);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_RTL);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_TTB_LTR);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_TTB_RTL);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_WEAK_LTR);
    PRINT_CONSTANT((gint) PANGO_DIRECTION_WEAK_RTL);
    PRINT_CONSTANT((gint) PANGO_ELLIPSIZE_END);
    PRINT_CONSTANT((gint) PANGO_ELLIPSIZE_MIDDLE);
    PRINT_CONSTANT((gint) PANGO_ELLIPSIZE_NONE);
    PRINT_CONSTANT((gint) PANGO_ELLIPSIZE_START);
    PRINT_CONSTANT((gint) PANGO_FONT_COLOR_DONT_CARE);
    PRINT_CONSTANT((gint) PANGO_FONT_COLOR_FORBIDDEN);
    PRINT_CONSTANT((gint) PANGO_FONT_COLOR_REQUIRED);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_COLOR);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_FAMILY);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_FEATURES);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_GRAVITY);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_SIZE);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_STRETCH);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_STYLE);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_VARIANT);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_VARIATIONS);
    PRINT_CONSTANT((guint) PANGO_FONT_MASK_WEIGHT);
    PRINT_CONSTANT((gint) PANGO_FONT_SCALE_NONE);
    PRINT_CONSTANT((gint) PANGO_FONT_SCALE_SMALL_CAPS);
    PRINT_CONSTANT((gint) PANGO_FONT_SCALE_SUBSCRIPT);
    PRINT_CONSTANT((gint) PANGO_FONT_SCALE_SUPERSCRIPT);
    PRINT_CONSTANT(PANGO_GLYPH_EMPTY);
    PRINT_CONSTANT(PANGO_GLYPH_INVALID_INPUT);
    PRINT_CONSTANT(PANGO_GLYPH_UNKNOWN_FLAG);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_AUTO);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_EAST);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_HINT_LINE);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_HINT_NATURAL);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_HINT_STRONG);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_NORTH);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_SOUTH);
    PRINT_CONSTANT((gint) PANGO_GRAVITY_WEST);
    PRINT_CONSTANT((guint) PANGO_LAYOUT_DESERIALIZE_CONTEXT);
    PRINT_CONSTANT((guint) PANGO_LAYOUT_DESERIALIZE_DEFAULT);
    PRINT_CONSTANT((gint) PANGO_LAYOUT_DESERIALIZE_INVALID);
    PRINT_CONSTANT((gint) PANGO_LAYOUT_DESERIALIZE_INVALID_VALUE);
    PRINT_CONSTANT((gint) PANGO_LAYOUT_DESERIALIZE_MISSING_VALUE);
    PRINT_CONSTANT((guint) PANGO_LAYOUT_SERIALIZE_CONTEXT);
    PRINT_CONSTANT((guint) PANGO_LAYOUT_SERIALIZE_DEFAULT);
    PRINT_CONSTANT((guint) PANGO_LAYOUT_SERIALIZE_OUTPUT);
    PRINT_CONSTANT((gint) PANGO_OVERLINE_NONE);
    PRINT_CONSTANT((gint) PANGO_OVERLINE_SINGLE);
    PRINT_CONSTANT((gint) PANGO_RENDER_PART_BACKGROUND);
    PRINT_CONSTANT((gint) PANGO_RENDER_PART_FOREGROUND);
    PRINT_CONSTANT((gint) PANGO_RENDER_PART_OVERLINE);
    PRINT_CONSTANT((gint) PANGO_RENDER_PART_STRIKETHROUGH);
    PRINT_CONSTANT((gint) PANGO_RENDER_PART_UNDERLINE);
    PRINT_CONSTANT(PANGO_SCALE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_AHOM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ANATOLIAN_HIEROGLYPHS);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ARABIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ARMENIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BALINESE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BASSA_VAH);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BATAK);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BENGALI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BOPOMOFO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BRAHMI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BRAILLE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BUGINESE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_BUHID);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CANADIAN_ABORIGINAL);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CARIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CAUCASIAN_ALBANIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CHAKMA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CHAM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CHEROKEE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_COMMON);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_COPTIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CUNEIFORM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CYPRIOT);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_CYRILLIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_DESERET);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_DEVANAGARI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_DUPLOYAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ELBASAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ETHIOPIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GEORGIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GLAGOLITIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GOTHIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GRANTHA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GREEK);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GUJARATI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_GURMUKHI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HANGUL);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HANUNOO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HATRAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HEBREW);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_HIRAGANA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_INHERITED);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_INVALID_CODE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KANNADA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KATAKANA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KAYAH_LI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KHAROSHTHI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KHMER);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KHOJKI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_KHUDAWADI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LAO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LATIN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LEPCHA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LIMBU);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LINEAR_A);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LINEAR_B);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LYCIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_LYDIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MAHAJANI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MALAYALAM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MANDAIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MANICHAEAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MENDE_KIKAKUI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MEROITIC_CURSIVE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MEROITIC_HIEROGLYPHS);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MIAO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MODI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MONGOLIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MRO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MULTANI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_MYANMAR);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_NABATAEAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_NEW_TAI_LUE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_NKO);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OGHAM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OLD_HUNGARIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OLD_ITALIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OLD_NORTH_ARABIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OLD_PERMIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OLD_PERSIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OL_CHIKI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_ORIYA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_OSMANYA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PAHAWH_HMONG);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PALMYRENE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PAU_CIN_HAU);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PHAGS_PA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PHOENICIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_PSALTER_PAHLAVI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_REJANG);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_RUNIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SAURASHTRA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SHARADA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SHAVIAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SIDDHAM);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SIGNWRITING);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SINHALA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SORA_SOMPENG);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SUNDANESE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SYLOTI_NAGRI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_SYRIAC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TAGALOG);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TAGBANWA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TAI_LE);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TAKRI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TAMIL);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TELUGU);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_THAANA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_THAI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TIBETAN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TIFINAGH);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_TIRHUTA);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_UGARITIC);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_UNKNOWN);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_VAI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_WARANG_CITI);
    PRINT_CONSTANT((gint) PANGO_SCRIPT_YI);
    PRINT_CONSTANT((guint) PANGO_SHAPE_NONE);
    PRINT_CONSTANT((guint) PANGO_SHAPE_ROUND_POSITIONS);
    PRINT_CONSTANT((guint) PANGO_SHOW_IGNORABLES);
    PRINT_CONSTANT((guint) PANGO_SHOW_LINE_BREAKS);
    PRINT_CONSTANT((guint) PANGO_SHOW_NONE);
    PRINT_CONSTANT((guint) PANGO_SHOW_SPACES);
    PRINT_CONSTANT((gint) PANGO_STRETCH_CONDENSED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_EXPANDED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_EXTRA_CONDENSED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_EXTRA_EXPANDED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_NORMAL);
    PRINT_CONSTANT((gint) PANGO_STRETCH_SEMI_CONDENSED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_SEMI_EXPANDED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_ULTRA_CONDENSED);
    PRINT_CONSTANT((gint) PANGO_STRETCH_ULTRA_EXPANDED);
    PRINT_CONSTANT((gint) PANGO_STYLE_ITALIC);
    PRINT_CONSTANT((gint) PANGO_STYLE_NORMAL);
    PRINT_CONSTANT((gint) PANGO_STYLE_OBLIQUE);
    PRINT_CONSTANT((gint) PANGO_TAB_CENTER);
    PRINT_CONSTANT((gint) PANGO_TAB_DECIMAL);
    PRINT_CONSTANT((gint) PANGO_TAB_LEFT);
    PRINT_CONSTANT((gint) PANGO_TAB_RIGHT);
    PRINT_CONSTANT((gint) PANGO_TEXT_TRANSFORM_CAPITALIZE);
    PRINT_CONSTANT((gint) PANGO_TEXT_TRANSFORM_LOWERCASE);
    PRINT_CONSTANT((gint) PANGO_TEXT_TRANSFORM_NONE);
    PRINT_CONSTANT((gint) PANGO_TEXT_TRANSFORM_UPPERCASE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_DOUBLE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_DOUBLE_LINE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_ERROR);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_ERROR_LINE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_LOW);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_NONE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_SINGLE);
    PRINT_CONSTANT((gint) PANGO_UNDERLINE_SINGLE_LINE);
    PRINT_CONSTANT((gint) PANGO_VARIANT_ALL_PETITE_CAPS);
    PRINT_CONSTANT((gint) PANGO_VARIANT_ALL_SMALL_CAPS);
    PRINT_CONSTANT((gint) PANGO_VARIANT_NORMAL);
    PRINT_CONSTANT((gint) PANGO_VARIANT_PETITE_CAPS);
    PRINT_CONSTANT((gint) PANGO_VARIANT_SMALL_CAPS);
    PRINT_CONSTANT((gint) PANGO_VARIANT_TITLE_CAPS);
    PRINT_CONSTANT((gint) PANGO_VARIANT_UNICASE);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_BOLD);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_BOOK);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_HEAVY);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_LIGHT);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_MEDIUM);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_NORMAL);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_SEMIBOLD);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_SEMILIGHT);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_THIN);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_ULTRABOLD);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_ULTRAHEAVY);
    PRINT_CONSTANT((gint) PANGO_WEIGHT_ULTRALIGHT);
    PRINT_CONSTANT((gint) PANGO_WRAP_CHAR);
    PRINT_CONSTANT((gint) PANGO_WRAP_NONE);
    PRINT_CONSTANT((gint) PANGO_WRAP_WORD);
    PRINT_CONSTANT((gint) PANGO_WRAP_WORD_CHAR);
    return 0;
}
